from datetime import date
import json
from typing import Optional

from ssb_altinn3_util.models.skjemadata.skjemadata_request_models import PeriodeRequestModel

from dapla_suv_tools._internals.integration import api_client
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util.validators import periode_id_validator, skjema_id_validator

PERIODE_PATH = "/skjemadata/periode"


@SuvOperationContext(validator=periode_id_validator)
def get_periode_by_id(*, periode_id: int, context: SuvOperationContext = None) -> OperationResult:
    try:
        content: str = api_client._get(path=f"{PERIODE_PATH}/{periode_id}", context=context)
        content_json = json.loads(content)
        context.log("info", "get_periode_by_id", f"Fetched periode with periode_id '{periode_id}'")

        return OperationResult(value=content_json, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to fetch for id {periode_id}", e)

        return OperationResult(success=False, value=context.errors(), log=context.logs())


@SuvOperationContext(validator=skjema_id_validator)
def get_perioder_by_skjema_id(*, skjema_id: int, context: SuvOperationContext = None) -> OperationResult:
    try:
        content: str = api_client._get(path=f"{PERIODE_PATH}/skjema/{skjema_id}", context=context)
        content_json = json.loads(content)
        context.log("info", "get_periode_by_skjema_id", f"Fetched perioder for skjema_id '{skjema_id}'")

        return OperationResult(value=content_json, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to fetch for skjema_id {skjema_id}", e)

        return OperationResult(success=False, value=context.errors(), log=context.logs())


@SuvOperationContext(validator=skjema_id_validator)
def create_periode(
        *,
        skjema_id: int,
        endret_av: str,
        periode_type: Optional[str] = None,
        periode_nr: Optional[int] = None,
        periode_aar: Optional[int] = None,
        periode_dato: Optional[date] = None,
        delreg_nr: Optional[int] = None,
        enhet_type: Optional[str] = None,
        vis_oppgavebyrde: Optional[str] = "N",
        vis_brukeropplevelse: Optional[str] = "N",
        altinn_tilgjengelig: Optional[date] = None,
        altinn_svarfrist: Optional[date] = None,
        context: SuvOperationContext = None
) -> OperationResult:
    model = PeriodeRequestModel(
        skjema_id=skjema_id,
        endret_av=endret_av,
        periode_type=periode_type,
        periode_nr=periode_nr,
        periode_aar=periode_aar,
        periode_dato=periode_dato,
        delreg_nr=delreg_nr,
        enhet_type=enhet_type,
        vis_oppgavebyrde=vis_oppgavebyrde,
        vis_brukeropplevelse=vis_brukeropplevelse,
        altinn_tilgjengelig=altinn_tilgjengelig,
        altinn_svarfrist=altinn_svarfrist
    )

    try:
        body = model.model_dump_json()
        content: str = api_client._post(path=PERIODE_PATH, body_json=body, context=context)
        new_id = json.loads(content)["id"]
        return OperationResult(value={"id": new_id}, log=context.logs())
    except Exception as e:
        context.set_error(
            f"Failed to create for skjema_id '{skjema_id}' - periode {periode_nr} {periode_type} {periode_nr}", e
        )
        return OperationResult(success=False, value=context.errors(), log=context.logs())


def delete_periode(*, periode_id: int, context: SuvOperationContext = None) -> OperationResult:
    try:
        content: str = api_client._delete(path=f"{PERIODE_PATH}/{periode_id}", context=context)
        return OperationResult(value=content, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to delete Periode with id '{periode_id}'.", e)
        return OperationResult(success=False, value=context.errors(), log=context.logs())



from datetime import date
from typing import Optional
import dapla_suv_tools._internals.client_apis.skjema_api as skjema_api
from dapla_suv_tools._internals.client_apis import periode_api
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util import constants
from dapla_suv_tools.pagination import PaginationInfo


class SuvClient:
    suppress_exceptions: bool
    operations_log: list

    def __init__(self, suppress_exceptions: bool = False):
        self.suppress_exceptions = suppress_exceptions
        self.operations_log = []

    def logs(self) -> list:
        return self.operations_log

    def flush_logs(self):
        self.operations_log = []

    # Skjema

    def get_skjema_by_id(self, *, skjema_id: int) -> dict:
        result = skjema_api.get_skjema_by_id(skjema_id=skjema_id)
        return self._process_result(result=result)

    def get_skjema_by_ra_nummer(
            self,
            *,
            ra_nummer: str,
            latest_only: bool = False,
            paginatipn_info: PaginationInfo = None) -> dict:
        result = skjema_api.get_skjema_by_ra_nummer(ra_nummer=ra_nummer, latest_only=latest_only, paging=paginatipn_info)
        return self._process_result(result=result)

    def create_skjema(
            self,
            *,
            ra_nummer: str,
            versjon: int,
            undersokelse_nr: str,
            gyldig_fra: date,
            endret_av: str,
            datamodell: Optional[str] = None,
            beskrivelse: Optional[str] = None,
            navn_nb: Optional[str] = None,
            navn_nn: Optional[str] = None,
            navn_en: Optional[str] = None,
            infoside: Optional[str] = None,
            eier: Optional[str] = None,
            kun_sky: bool = False,
            gyldig_til: Optional[date] = None,
    ) -> dict:
        fwd_args = {k: v for k,v in locals().items() if k not in {"self", "*"}}

        result = skjema_api.create_skjema(**fwd_args)
        return self._process_result(result=result)

    def delete_skjema(self, skjema_id: int) -> dict:
        result = skjema_api.delete_skjema(skjema_id=skjema_id)
        return self._process_result(result=result)

    # Periode

    def get_periode_by_id(self, *, periode_id: int) -> dict:
        result = periode_api.get_periode_by_id(periode_id=periode_id)
        return self._process_result(result=result)

    def get_perioder_by_skjema_id(self, *, skjema_id: int) -> dict:
        result = periode_api.get_perioder_by_skjema_id(skjema_id=skjema_id)
        return self._process_result(result=result)

    def create_periode(
            self,
            *,
            skjema_id: int,
            endret_av: str,
            periode_type: Optional[str] = None,
            periode_nr: Optional[int] = None,
            periode_aar: Optional[int] = None,
            periode_dato: Optional[date] = None,
            delreg_nr: Optional[int] = None,
            enhet_type: Optional[str] = None,
            vis_oppgavebyrde: Optional[str] = "N",
            vis_brukeropplevelse: Optional[str] = "N",
            altinn_tilgjengelig: Optional[date] = None,
            altinn_svarfrist: Optional[date] = None,
    ) -> dict:
        fwd_args = {k: v for k, v in locals().items() if k not in {"self", "*"}}
        result = periode_api.create_periode(**fwd_args)
        return self._process_result(result=result)

    def delete_periode(self, *, periode_id: int) -> dict:
        result = periode_api.delete_periode(periode_id=periode_id)
        return self._process_result(result=result)

    def _process_result(self, result: OperationResult) -> dict:
        self.operations_log.append(result.operation_log)
        if result.result == constants.OPERATION_OK:
            return result.result_json

        if result.result == constants.OPERATION_ERROR:
            if self.suppress_exceptions:
                return result.result_json
            errors = result.result_json["errors"]
            raise errors[len(errors) - 1]["exception"]

        return {"result": "Undefined result.  This shouldn't happen."}



from functools import wraps
from typing import Any, Callable, Optional
import inspect


class SuvOperationContext:

    state: dict[str, Any]
    func_kwargs: dict[str, Any]
    arg_validator: Callable[["SuvOperationContext", dict], None]

    def __init__(self, validator: Callable[["SuvOperationContext", dict], None] = None, func_kwargs: dict[str, Any] = None):
        self.arg_validator = validator
        self._flush_state()
        self.func_kwargs = func_kwargs

    def __enter__(self):
        self._flush_state()
        try:
            if self.func_kwargs and self.arg_validator:
                self.arg_validator(self, **self.func_kwargs)

            return self
        except Exception as e:
            self.__exit__(type(e), e, e.__traceback__)
            raise

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        if exc_type is not None:
            if self.state["errors"]:
                print("Errors flagged during operation:")
                for error in self.state["errors"]:
                    print(error)
            return False
        return True

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.log("info", func.__name__, f"Starting function call '{func.__name__}'", result="N/A")
            if self.arg_validator is not None:
                self.arg_validator(self, **kwargs)
            self.func_args = args
            self.func_kwargs = kwargs
            with self._recreate_context_manager():
                sig = inspect.signature(func)
                if "context" in sig.parameters or "kwargs" in sig.parameters:
                    return func(*args, **kwargs, context=self)
                return func(*args, **kwargs)

        return wrapper

    def _recreate_context_manager(self):
        return self

    def log(self, level: str, operation: str, message: str = "", result: str = "OK"):
        self.state["log"].append({
            "level": level,
            "operation": operation,
            "message": message,
            "result": result
        })

    def set_error(self, error_msg: str, exception: Exception):
        self.state["errors"].append({
            "error_type": type(exception).__name__,
            "error_message": error_msg,
            "exception": exception
        })

    def errors(self) -> dict:
        return {"errors": self.state["errors"]}

    def logs(self) -> dict:
        return {"logs": self.state["log"]}

    def _flush_state(self):
        self.state = {
            "log": [],
            "errors": []
        }


import inspect
import os

import requests

from dapla.auth import AuthClient

from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext

END_USER_API_BASE_URL = os.getenv("SUV_END_USER_API_URL")


def _get(path: str, context: SuvOperationContext) -> str:
    headers = _get_headers()

    response = requests.get(f"{END_USER_API_BASE_URL}{path}", headers=headers)

    return _handle_response(response=response, context=context)


def _post(path: str, body_json: str, context: SuvOperationContext) -> str:
    headers = _get_headers()

    response = requests.post(url=f"{END_USER_API_BASE_URL}{path}", headers=headers, data=body_json)

    return _handle_response(response=response, context=context)


def _delete(path: str, context: SuvOperationContext) -> str:
    headers = _get_headers()

    response = requests.delete(url=f"{END_USER_API_BASE_URL}{path}", headers=headers)

    return _handle_response(response, context=context)


def _handle_response(response: requests.Response, context: SuvOperationContext) -> str:

    called = _get_caller(2)
    caller = _get_caller(3)

    msg = f"calling '{called}' from '{caller}'."

    if not _success(response.status_code):
        error = response.content.decode("UTF-8")
        ex = Exception(f"Failed call to api while {msg}.")
        context.set_error(f"Error (status: {response.status_code}) {msg}:  {error}", ex)
        raise ex

    context.log("info", "api_client.POST", msg)

    return response.content.decode("UTF-8")


def _get_headers() -> dict:
    token: str = AuthClient.fetch_personal_token()

    return {
        "authorization": f"Bearer {token}",
        "content-type": "application/json"
    }


def _get_caller(depth: int) -> str:
    frames = inspect.stack()
    caller = frames[depth]
    return caller.function


def _success(status_code: int) -> bool:
    return str(status_code).startswith("2")
